#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from . import chart_object
from . import color
from . import fill_style_doc
from . import line_style
from . import object_set
from . import pychart_util
from .pychart_types import *
from .scaling import *

_keys = {
    "bgcolor": (color.T, color.white, "The background color."),
    "line_style": (line_style.T, line_style.default,
                   pychart_util.line_desc),
    "line_interval": (NumberType, 3,
                      "The interval between successive stitch lines.")
}


class T(chart_object.T):
    __doc__ = fill_style_doc.doc
    keys = _keys

    ##AUTOMATICALLY GENERATED
    ##END AUTOMATICALLY GENERATED
    def __str__(self):
        s = name_table().lookup(self)
        if s:
            return s
        return "<fillstyle: bg=%s line=%s interval=%s>" % \
               (self.bgcolor, self.line_style, self.line_interval)


class Plain(T):
    """This class just fills the region with solid background color.
Attributes line_style and line_interval are ignored."""

    def draw(self, can, x1, y1, x2, y2):
        pass


class Diag(T):
    "This class fills the region with diagonal lines."

    def draw(self, can, x1, y1, x2, y2):
        line_width = self.line_style.width
        interval = self.line_interval * 1.414
        x1 -= line_width
        y1 -= line_width
        x2 += line_width
        y2 += line_width
        length = max(y2 - y1, x2 - x1)
        curx = x1 - length
        while curx < x2:
            can.line(self.line_style, curx, y1, curx + length, y1 + length)
            curx += interval


class Rdiag(T):
    """Fills the region with diagonal lines, but tilted in the opposite
direction from fill_style.Diag."""

    def draw(self, can, x1, y1, x2, y2):
        line_width = self.line_style.width
        interval = self.line_interval * 1.414
        x1 -= line_width
        y1 -= line_width
        x2 += line_width
        y2 += line_width
        length = max(y2 - y1, x2 - x1)
        curx = x1
        while curx < x2 + length:
            can.line(self.line_style, curx, y1, curx - length, y1 + length)
            curx += interval


class Vert(T):
    "Fills the region with vertical lines"

    def draw(self, can, x1, y1, x2, y2):
        interval = self.line_interval
        curx = x1
        while curx < x2:
            can.line(self.line_style, curx, y1, curx, y2)
            curx += interval


class Horiz(T):
    "Fills the region with horizontal lines"

    def draw(self, can, x1, y1, x2, y2):
        interval = self.line_interval
        cury = y1
        while cury < y2:
            can.line(self.line_style, x1, cury, x2, cury)
            cury += interval


class Stitch(T):
    "Fills the region with horizontal and vertical lines."

    def draw(self, can, x1, y1, x2, y2):
        interval = self.line_interval
        cury = y1
        while cury < y2:
            can.line(self.line_style, x1, cury, x2, cury)
            cury += interval
        curx = x1
        while curx < x2:
            can.line(self.line_style, curx, y1, curx, y2)
            curx += interval


class Wave(T):
    "Fills the region with horizontal wavy lines."

    def draw(self, can, x1, y1, x2, y2):
        x1 = xscale(x1)
        x2 = xscale(x2)
        y1 = yscale(y1)
        y2 = yscale(y2)
        line_width = nscale(self.line_style.width)
        interval = nscale(self.line_interval)

        can.set_line_style(self.line_style)
        x1 -= line_width
        x2 += line_width
        cury = y1
        half = interval / 2.0
        while cury < y2:
            curx = x1
            can.newpath()
            can.moveto(curx, cury)
            while curx < x2:
                can.lineto(curx + half, cury + half)
                can.lineto(curx + interval, cury)
                curx += interval
            can.stroke()
            cury += interval


class Vwave(T):
    """Fills the region with vertical wavy lines."""

    def draw(self, can, x1, y1, x2, y2):
        x1 = xscale(x1)
        x2 = xscale(x2)
        y1 = yscale(y1)
        y2 = yscale(y2)
        line_width = nscale(self.line_style.width)
        interval = nscale(self.line_interval)

        can.set_line_style(self.line_style)
        y1 -= line_width
        y2 += line_width
        curx = x1
        half = interval / 2.0
        while curx < x2:
            cury = y1
            can.newpath()
            can.moveto(curx, cury)
            while cury < y2:
                can.lineto(curx + half, cury + half)
                can.lineto(curx, cury + interval)
                cury += interval
            can.stroke()
            curx += interval


class Lines(T):
    """Fills the region with a series of short line segments."""

    def draw(self, can, x1, y1, x2, y2):
        interval = nscale(self.line_interval)
        cury = y1
        j = 0
        while cury < y2:
            curx = x1
            if j % 2 == 1:
                curx += interval / 2.0
            while curx < x2:
                can.line(self.line_style, curx, cury, curx + interval / 2.0, cury)
                curx += interval * 1.5
            j += 1
            cury += interval


default = Plain()

color_standards = object_set.T()
grayscale_standards = object_set.T()


def _intern_both(style):
    global color_standards, grayscale_standards
    color_standards.add(style)
    grayscale_standards.add(style)
    return style


def _intern_color(style):
    global color_standards, grayscale_standards
    color_standards.add(style)
    return style


def _intern_grayscale(style):
    global color_standards, grayscale_standards
    grayscale_standards.add(style)
    return style


black = _intern_both(Plain(bgcolor=color.gray_scale(0.0), line_style=None))

red = _intern_color(Plain(bgcolor=color.red))
darkseagreen = _intern_color(Plain(bgcolor=color.darkseagreen))
blue = _intern_color(Plain(bgcolor=color.blue))
aquamarine1 = _intern_color(Plain(bgcolor=color.aquamarine1))
gray70 = _intern_both(Plain(bgcolor=color.gray70, line_style=None))
brown = _intern_color(Plain(bgcolor=color.brown))
darkorchid = _intern_color(Plain(bgcolor=color.darkorchid))
diag = _intern_both(Diag(line_style=line_style.T(cap_style=2)))
green = _intern_color(Plain(bgcolor=color.green))
gray50 = _intern_both(Plain(bgcolor=color.gray50, line_style=None))
white = _intern_both(Plain(bgcolor=color.gray_scale(1.0), line_style=None))
goldenrod = _intern_color(Plain(bgcolor=color.goldenrod))
rdiag = _intern_both(Rdiag(line_style=line_style.T(cap_style=2)))
vert = _intern_both(Vert(line_interval=1.8))

gray30 = _intern_both(Plain(bgcolor=color.gray30, line_style=None))
gray20 = _intern_both(Plain(bgcolor=color.gray20, line_style=None))
gray10 = _intern_both(Plain(bgcolor=color.gray10, line_style=None))
diag2 = _intern_both(Diag(line_style=line_style.T(width=3, cap_style=2),
                          line_interval=6))
rdiag2 = _intern_both(Rdiag(line_style=line_style.T(width=3, cap_style=2),
                            line_interval=6))
yellow = _intern_color(Plain(bgcolor=color.yellow))
diag3 = _intern_both(Diag(line_style=line_style.T(width=3, color=color.gray50, cap_style=2),
                          line_interval=6))
horiz = _intern_both(Horiz(line_interval=1.8))
gray90 = _intern_both(Plain(bgcolor=color.gray90, line_style=None))
rdiag3 = _intern_both(Rdiag(line_style=line_style.T(width=3,
                                                    color=color.gray50,
                                                    cap_style=2),
                            line_interval=6))

wave = _intern_both(Wave(line_style=line_style.T(cap_style=2, join_style=1)))
vwave = _intern_both(Vwave(line_style=line_style.T(cap_style=2, join_style=1)))
stitch = _intern_both(Stitch(line_style=line_style.T(cap_style=2, join_style=1)))
lines = _intern_both(Lines(line_style=line_style.T()))

diag_fine = _intern_both(Diag(line_style=line_style.T(width=0.75, cap_style=2),
                              line_interval=1.5))
diag2_fine = _intern_both(Diag(line_style=line_style.T(width=0.75, cap_style=2),
                               line_interval=1.5))
diag3_fine = _intern_both(Diag(line_style=line_style.T(width=0.75,
                                                       color=color.gray50,
                                                       cap_style=2),
                               line_interval=1.5))
rdiag_fine = _intern_both(Rdiag(line_style=line_style.T(width=0.75, cap_style=2),
                                line_interval=1.5))
rdiag2_fine = _intern_both(Rdiag(line_style=line_style.T(width=0.75, cap_style=2),
                                 line_interval=1.5))
rdiag3_fine = _intern_both(Rdiag(line_style=line_style.T(width=0.75,
                                                         color=color.gray50,
                                                         cap_style=2),
                                 line_interval=1.5))

horiz_fine = _intern_both(Horiz(line_interval=1.5))
vert_fine = _intern_both(Vert(line_interval=1.5))

#
# Fill styles for color charts.
#

standards = None
_name_table = None


def init():
    global standards, _name_table
    if theme.use_color:
        standards = color_standards
    else:
        standards = grayscale_standards
    _name_table = None


def name_table():
    global _name_table
    if not _name_table:
        _name_table = pychart_util.symbol_lookup_table(globals(), standards)
    return _name_table


init()
theme.add_reinitialization_hook(init)



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import math

from . import arrow_doc
from . import canvas
from . import chart_object
from . import color
from . import line_style
from . import object_set
from .pychart_types import *
from .scaling import *

__doc__ = """
Arrow is an optional component of a chart that draws line segments with
an arrowhead. To draw an arrow, one creates an arrow.T object, and calls
its "draw" method usually after area.draw() is called (otherwise, area.draw()
may overwrite the arrow). For example, the below code draws an arrow
from (10,10) to (20,30).

ar = area.T(...)
a = arrow.T(head_style = 1)
ar.draw()
a.draw([(10,10), (20,30)])
"""


def draw_arrowhead(can, tailx, taily, tipx, tipy, thickness, head_len, style):
    can.comment("ARROWHEAD tail=(%d,%d) tip=(%d,%d)\n"
                % (tailx, taily, tipx, tipy))

    halfthickness = thickness / 2.0
    dx = tipx - tailx
    dy = tipy - taily
    arrow_len = math.sqrt(dx * dx + dy * dy)
    angle = math.atan2(dy, dx) * 360 / (2 * math.pi)
    base = arrow_len - head_len
    can.push_transformation((tailx, taily), None, angle)

    can.newpath()
    if style == 0:
        can.moveto(base, - halfthickness)
        can.lineto(base, halfthickness)
        can.lineto(arrow_len, 0)
        can.closepath()
    elif style == 1:
        depth = head_len / 2.5
        can.moveto(base - depth, -halfthickness)
        can.lineto(base, 0)
        can.lineto(base - depth, halfthickness)
        can.lineto(arrow_len, 0)
        can.closepath()
    elif style == 2:
        can.moveto(base + head_len / 2.0, 0)
        can.path_arc(base + head_len / 2.0, 0, head_len / 2.0, 1.0, 0, 400)
    elif style == 3:
        can.moveto(base, 0)
        can.lineto(base + head_len / 2.0, -halfthickness)
        can.lineto(arrow_len, 0)
        can.lineto(base + head_len / 2.0, halfthickness)
        can.closepath()
    else:
        raise Exception("Arrow style must be a number between 0 and 3.")
    can.fill()
    can.pop_transformation()
    can.comment("end ARROWHEAD.\n")


def draw_arrowbody(can, tailx, taily, tipx, tipy, head_len):
    dx = tipx - tailx
    dy = tipy - taily
    arrow_len = math.sqrt(dx * dx + dy * dy)
    angle = math.atan2(dy, dx) * 360 / (2 * math.pi)
    base = arrow_len - head_len
    can.push_transformation((tailx, taily), None, angle)
    can.moveto(0, 0)
    can.lineto(base + head_len * 0.1, 0)
    can.stroke()
    can.pop_transformation()


class T(chart_object.T):
    __doc__ = arrow_doc.doc
    keys = {
        "thickness": (UnitType, 4,
                      "The width of the arrow head."),
        "head_len": (UnitType, 8,
                     "The length of the arrow head."),
        "head_color": (color.T, color.default,
                       "The color of the arrow head."),
        "line_style": (line_style.T, line_style.default,
                       "Line style."),
        "head_style": (IntType, 1,
                       "The value of 0 draws a triangular arrow head. The value of 1 draws a swallow-tail arrow head. The value of 2 draws a circular head. The value of 3 draws a diamond-shaped head.")
    }

    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED
    def draw(self, points, can=None):
        """Parameter <points> specifies the
        list of points the arrow traverses through.
        It should contain at least two points, i.e.,
        the tail and tip. Parameter
        <can> is an optional parameter that specifies the output.
        <<canvas>>
        """
        if can == None: can = canvas.default_canvas()
        assert self.check_integrity()
        xtip = points[-1][0]
        ytip = points[-1][1]

        xtail = points[-2][0]
        ytail = points[-2][1]

        can.newpath()
        can.set_line_style(self.line_style)
        if len(points) > 2:
            can.moveto(points[0][0], points[0][1])
            for i in range(1, len(points) - 1):
                can.lineto(points[i][0], points[i][1])

        draw_arrowbody(can, xscale(xtail), yscale(ytail),
                       yscale(xtip), yscale(ytip),
                       nscale(self.head_len))

        can.set_fill_color(self.head_color)
        draw_arrowhead(can, xscale(xtail), yscale(ytail),
                       xscale(xtip), yscale(ytip),
                       nscale(self.thickness),
                       nscale(self.head_len),
                       self.head_style)

        can.setbb(xtail, ytail)
        can.setbb(xtip, ytip)


standards = object_set.T()


def _intern(a):
    global standards
    standards.add(a)
    return a


a0 = _intern(T(head_style=0))
a1 = _intern(T(head_style=1))
a2 = _intern(T(head_style=2))
a3 = _intern(T(head_style=3))
gray0 = _intern(T(head_style=0, head_color=color.gray50,
                  line_style=line_style.T(color=color.gray50)))
gray1 = _intern(T(head_style=1, head_color=color.gray50,
                  line_style=line_style.T(color=color.gray50)))
gray2 = _intern(T(head_style=2, head_color=color.gray50,
                  line_style=line_style.T(color=color.gray50)))
gray3 = _intern(T(head_style=3, head_color=color.gray50,
                  line_style=line_style.T(color=color.gray50)))

fat0 = _intern(T(head_style=0, head_len=12, thickness=10, line_style=line_style.T(width=2)))
fat1 = _intern(T(head_style=1, head_len=12, thickness=10, line_style=line_style.T(width=2)))
fat2 = _intern(T(head_style=2, head_len=12, thickness=10, line_style=line_style.T(width=2)))
fat3 = _intern(T(head_style=3, head_len=12, thickness=10, line_style=line_style.T(width=2)))
fatgray0 = _intern(T(head_style=0, head_len=12, thickness=10,
                     head_color=color.gray50,
                     line_style=line_style.T(width=2, color=color.gray50)))
fatgray1 = _intern(T(head_style=1, head_len=12, thickness=10,
                     head_color=color.gray50,
                     line_style=line_style.T(width=2, color=color.gray50)))
fatgray2 = _intern(T(head_style=2, head_len=12, thickness=10,
                     head_color=color.gray50,
                     line_style=line_style.T(width=2, color=color.gray50)))
fatgray3 = _intern(T(head_style=3, head_len=12, thickness=10,
                     head_color=color.gray50,
                     line_style=line_style.T(width=2, color=color.gray50)))

default = a1



# -*- coding: utf-8 -*-
_tdl_no_reimport = True

import os.path
import time
from html.parser import HTMLParser

import Kittens.utils

import Purr

_verbosity = Kittens.utils.verbosity(name="purrparse")
dprint = _verbosity.dprint
dprintf = _verbosity.dprintf


class LogIndexParser(HTMLParser):
    def reset(self):
        HTMLParser.reset(self)
        self.title = None
        self.timestamp = None
        self.curclass = None

    def end(self):
        dprintf(4, "end, title '%s', timestamp %s", self.title, self.timestamp)

    def handle_starttag(self, tag, attrs):
        dprint(4, "start tag", tag, attrs)
        # anchor tag -- data we look for is in here
        if tag == "a":
            attrs = dict(attrs)
            tagclass = attrs.get('class')
            if tagclass:
                # whenever we encounter an A tag with a CLASS attribute, and the class
                # has a _handle_start_class method defined, call the handler.
                # The attributes are passed as keywords to the handler
                start_handler = getattr(self, "_handle_start_%s" % tagclass, None)
                if callable(start_handler):
                    start_handler(**attrs)
                # If the class also has a _handle_end_class method defined, accumulate all text
                # inside the tag (in self.curdata) for handling in the end handler
                if hasattr(self, "_handle_end_%s" % tagclass):
                    if self.curclass:
                        raise ValueError("nested class %s inside tag of class %s" % (tagclass, self.curclass))
                    self.curclass = tagclass
                    self.curdata = ""
        # paragraph tag: add newline to curdata, if accumulating
        elif tag == "p":
            if self.curclass and self.curdata:
                self.curdata += "\n"
        elif tag == "br":
            if self.curclass and self.curdata:
                self.curdata += "<BR>"

    def _handle_start_TITLE(self, timestamp=0, **kw):
        if self.title is None:
            try:
                self.timestamp = int(float(timestamp))
            except:
                self.timestamp = time.time()

    def _handle_end_TITLE(self, data):
        if self.title is None:
            self.title = data

    def handle_data(self, data):
        dprintf(4, "data: {%s}\n", data)
        # if curclass is None, we're not accumulating data, just skip it
        if self.curclass is None:
            return
        # is there anything here except whitespace? Append to data, but
        # replace newlines with spaces
        if data.rstrip():
            self.curdata += data.replace("\n", " ")
        # else all space. Append a single space to curdata, if it doesn't
        # already end in a space
        else:
            if self.curdata and self.curdata[-1] not in "\n ":
                self.curdata += " "

    _entity_dict = dict(lt="<", gt=">")

    def handle_entityref(self, name):
        dprintf(4, "entityref: {%s}\n", name)
        data = self._entity_dict.get(name, None)
        if data:
            self.handle_data(data)

    def handle_endtag(self, tag):
        dprint(4, "end tag", tag)
        # if end of A tag with CLASS attribute, pass accumulated data to data handler
        if tag == "a":
            if self.curclass:
                getattr(self, "_handle_end_%s" % self.curclass)(self.curdata)
                self.curclass = None
        elif tag == "html":
            self.end()


class LogEntryIndexParser(LogIndexParser):
    def __init__(self, dirname):
        LogIndexParser.__init__(self)
        self._dirname = dirname

    def reset(self):
        LogIndexParser.reset(self)
        self.comments = None
        self.dps = []
        self._new_dp = None

    def end(self):
        self._add_data_product()
        LogIndexParser.end(self)

    def _handle_start_DP(self, filename=None, src=None, policy=None, quiet=False,
                         timestamp=0, comment=None, render=None, **kw):
        # dispence with previous DP tag, if any
        self._add_data_product()
        # setup data for this tag
        comment = comment or ""
        try:
            timestamp = float(timestamp)
        except:
            timestamp = time.time()
        if not isinstance(quiet, bool):
            try:
                quiet = bool(int(quiet))
            except:
                quiet = bool(quiet)
        comment = comment.replace("&lt;", "<").replace("&gt;", ">")
        self._new_dp = Purr.DataProduct(filename=filename, sourcepath=src,
                                        timestamp=timestamp, comment=comment,
                                        fullpath=os.path.join(self._dirname, filename or ""),
                                        policy=policy, render=render, quiet=quiet, archived=True)

    def _handle_end_TITLE(self, data):
        self.title = data.replace("&lt;", "<").replace("&gt;", ">")

    def _handle_end_COMMENTS(self, data):
        self.comments = data.replace("&lt;", "<").replace("&gt;", ">")

    def _handle_end_DPCOMMENT(self, data):
        if self._new_dp:
            self._new_dp.comment = data.replace("&lt;", "<").replace("&gt;", ">")
            self._add_data_product()

    def _add_data_product(self):
        if self._new_dp:
            self.dps.append(self._new_dp)
            self._new_dp = None



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from . import arrow
from . import canvas
from . import chart_object
from . import fill_style
from . import font
from . import line_style
from . import text_box_doc
from . import pychart_util
from .pychart_types import *


class T(chart_object.T):
    __doc__ = text_box_doc.doc
    keys = {"text": (StringType, "???", "Text body. <<font>>"),
            "loc": (TupleType, (0, 0),
                    "The location of the text box."),
            "line_style": (line_style.T, line_style.default,
                           """The line style of the surrounding frame."""),
            "fill_style": (fill_style.T, fill_style.white,
                           "Specifies the fill style of the text box."),
            "top_fudge": (UnitType, 0,
                          "The amount of space (in points) above the first line"),
            "bottom_fudge": (UnitType, 5,
                             "The amount of space below the last line"),
            "left_fudge": (UnitType, 5,
                           "The amount of space left of the box"),
            "right_fudge": (UnitType, 5,
                            "The amount of space right of the box"),
            "_arrows": (ListType, pychart_util.new_list, "The list of arrows. Not to be touched by the user directly"),
            "radius": (UnitType, 0,
                       """Radius of the four corners of the rectangle.
                       If the value is zero, a sharp-cornered
                       rectangle is drawn."""),
            "shadow": (ShadowType, None,
                       pychart_util.shadow_desc)
            }

    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED
    def get_dimension(self):
        x = self.loc[0] - self.left_fudge
        y = self.loc[1] - self.bottom_fudge
        width = font.text_width(self.text) + self.right_fudge + self.left_fudge
        height = (font.text_height(self.text))[0] + self.top_fudge + self.bottom_fudge
        return (x, y, width, height)

    def choose_end_point(self, tipx, tipy):
        (x, y, width, height) = self.get_dimension()

        minDist = -1
        minPoint = None
        vertices = [(x, y),
                    (x + width, y),
                    (x + width, y + height),
                    (x, y + height)]

        if tipx >= x and tipx < x + width:
            vertices.append((tipx, y))
            vertices.append((tipx, y + height))
        if tipy >= y and tipy < y + height:
            vertices.append((x, tipy))
            vertices.append((x + width, tipy))

        for startPoint in vertices:
            dist = ((startPoint[0] - tipx) ** 2 + (startPoint[1] - tipy) ** 2)
            if not minPoint or dist < minDist:
                minPoint = startPoint
                minDist = dist

        return minPoint

    def add_arrow(self, tipLoc, tail=None, arrow=arrow.default):
        """This method adds a straight arrow that points to
        @var{TIPLOC}, which is a tuple of integers. @var{TAIL}
        specifies the starting point of the arrow. It is either None
        or a string consisting of the following letters: 'l', 'c',
        'r', 't', 'm,', and 'b'.  Letters 'l', 'c', or 'r' means to
        start the arrow from the left, center, or right of the text
        box, respectively. Letters 't', 'm', or 'b' means to start the
        arrow from the top, middle or bottom of the text box.  For
        example, when @samp{tail = 'tc'} then arrow is drawn from
        top-center point of the text box. ARROW specifies the style of
        the arrow. <<arrow>>.
        """
        self._arrows.append((tipLoc, tail, arrow))

    def draw(self, can=None):
        if can == None:
            can = canvas.default_canvas()
        x = self.loc[0]
        y = self.loc[1]
        text_width = font.text_width(self.text)
        text_height = font.text_height(self.text)[0]
        (halign, valign, angle) = font.get_align(self.text)

        if self.line_style or self.fill_style:
            width = text_width + self.left_fudge + self.right_fudge
            height = text_height + self.bottom_fudge + self.top_fudge
            can.round_rectangle(self.line_style, self.fill_style,
                                x - self.left_fudge, y - self.bottom_fudge,
                                x - self.left_fudge + width, y - self.bottom_fudge + height,
                                self.radius, self.shadow)

        if halign == 'L':
            can.show(x, y, self.text)
        elif halign == 'C':
            can.show(x + text_width / 2.0, y, self.text)
        elif halign == 'R':
            can.show(x + text_width, y, self.text)
        else:
            raise Exception("Unsupported alignment (" + halign + ")")

        # draw arrows
        for t in self._arrows:
            (tipLoc, tail, arrow) = t
            if tail:
                (x, y, width, height) = self.get_dimension()
                origin = [x, y]
                for ch in tail:
                    if ch == 'l':
                        origin[0] = x
                    elif ch == 'c':
                        origin[0] = x + width / 2.0
                    elif ch == 'r':
                        origin[0] = x + width
                    elif ch == 'b':
                        origin[1] = y
                    elif ch == 'm':
                        origin[1] = y + height / 2.0
                    elif ch == 't':
                        origin[1] = y + height
                    else:
                        raise ValueError(tail + ": unknown tail location spec.")
            else:
                origin = self.choose_end_point(tipLoc[0], tipLoc[1])
            arrow.draw((origin, tipLoc), can)



# -*- coding: utf-8 -*-

import os
import os.path
import time
import traceback
from past.builtins import cmp

import Kittens.config
import Kittens.utils

from Purr import ConfigFile

_verbosity = Kittens.utils.verbosity(name="render")
dprint = _verbosity.dprint
dprintf = _verbosity.dprintf

from future.standard_library import hooks

with hooks():
    from six.moves.urllib.parse import quote as quote_url


def renderDefault(dp, relpath):
    """Fall-back rendering method (if all else fails), renders DP as a link"""
    return """<A HREF="%s">%s</A>""" % (quote_url(relpath + dp.filename), dp.filename)


class DefaultRenderer(object):
    # this gives a short ID for the class (used in GUIs and such)
    renderer_id = "link"

    # this gives a documentation string. You can use rich text here
    renderer_doc = """<P>The "link" renderer shows data products as simple HTML links.<P>"""

    # this is set (by addRenderer() below) to the file modification time of the module containing
    # the renderer. Decisions on whether to re-render costly products are made based on this mtime.
    module_mtime = 0

    """This is a base class for data product rendering plugins."""

    @staticmethod
    def canRender(filename):
        """The canRender() method of a renderer class is called for every new data product.
        It determines whether the DP can be rendered by this class.
        It should return False if the DP cannot be rendered, or else a number giving the
        priority of the renderer. Lower-priority renderers will be offered to the user
        as defaults."""
        return 1000000


    def __init__(self, dp, refresh=0):
        """A renderer is initialized with a LogEntry.DataProduct object.
        If a renderer has been initialized, it is guaranteed to be used, so any "heavy" activity
        such as making of thumbnails, etc. can already be undertaken here.
        If refresh is set to a timestamp, then any subproducts (thumbnails, HTML caches, etc.) older than the timestamp will need to be regenerated.
        """
        self.dp = dp
        # setup some convenient attributes
        # filename: base filename of DP (w/o path)
        self.filename = dp.filename
        # dirname: directory of DP
        self.dirname = os.path.dirname(dp.fullpath)
        # base name of DP, without extension
        self.basename = os.path.splitext(dp.filename)[0]
        # base path to DP, without extension
        self.basepath = os.path.splitext(dp.fullpath)[0]
        # modification time of DP
        file_mtime = os.path.getmtime(dp.fullpath)
        self.file_mtime = max(file_mtime, refresh)
        dprintf(3, "created renderer for %s. File timestamp is %s, refresh cutoff is %s\n",
                dp.fullpath,
                time.strftime("%x %X", time.localtime(file_mtime)),
                time.strftime("%x %X", time.localtime(refresh)))

    def subproductPath(self, ext):
        """Makes a subproduct filename by appending 'ext' to the subproduct directory.
        Returns a (filename,fullpath) tuple."""
        ext = ext.lstrip("-")
        basename = os.path.join(os.path.basename(self.dp.subproduct_dir()), ext)
        fullpath = os.path.join(self.dp.subproduct_dir(), ext)
        return basename, fullpath

    def subproductUpToDate(self, fpath):
        """Returns True if the path is up-to-date w.r.t. our data product: i.e. exists, and is no older
        than the DP itself, or the module itself (the last check ensures that subproducts are remade
        if the rendering code changes.)"""
        # subproduct is up-to-date if it is older than the file mtime
        if not os.path.exists(self.dp.subproduct_dir()):
            os.mkdir(self.dp.subproduct_dir())
        ## This caused many re-rendering whenever a plugin got recompiled, so I disabled it.
        ## Need to add a control for re-rendering.
        ##    if os.path.exists(fpath) and os.path.getmtime(fpath) >= max(self.file_mtime,self.module_mtime):
        if os.path.exists(fpath) and os.path.getmtime(fpath) >= self.file_mtime:
            dprintf(3, "subproduct %s is up-to-date, no need to regenerate\n", fpath)
            return True
        else:
            if not os.path.exists(fpath):
                dprintf(3, "subproduct %s does not yet exist, will generate\n", fpath)
            else:
                dprintf(4, "subproduct %s is out of date (timestamp %s, file %s, module %s), will regenerate\n", fpath,
                        time.strftime("%x %X", time.localtime(os.path.getmtime(fpath))),
                        time.strftime("%x %X", time.localtime(self.file_mtime)),
                        time.strftime("%x %X", time.localtime(self.module_mtime)))
            return False

    def subproduct(self, ext):
        """Makes a subproduct filename by calling subproductPath(), then calls
        subproductUpToDate() to determine if it is up-to-date.
        Returns tuple of basename,path,uptodate
        """
        fname, fpath = self.subproductPath(ext)
        return fname, fpath, self.subproductUpToDate(fpath)

    # This is a dict of renderer-specific options.
    options = {} # type: dict
    # Its life is a bit interesting because:
    #   1. If subclass X adds new options (via addOption() below), this dict is copied
    #     and reassigned to X.options, so X has its own class-specific copy
    #   2. If subclass Y (subclass of X) add yet more options, the dict is copied again.
    #     so now both X and Y have class-specific copies.
    #   3. If y is an instance of Y, and y.setOption() is called, the dict is copied again
    #     and reassigned to y.options, so y has its own instance-specific copy.
    # _options_owner is set to the owner of the current dict. It is initially DefaultRenderer,
    # then in case (1) it is X, in (2) it is Y, and in (3) it is y itself.
    _options_owner = None

    @classmethod
    def addOption(classobj, name, default, dtype=str, doc=None):
        """Adds a renderer option named 'name', with the given default value.
        'dtype' must be a callable to convert a string to an option.
        'doc' is a doc string.
        Options will be initialized from config file here.
        """
        # make a config object
        if not hasattr(classobj, '_config'):
            classobj._config = Kittens.config.SectionParser(ConfigFile, "render-" + classobj.renderer_id)
        # make a class-specific copy of the current option set
        if classobj._options_owner is not classobj:
            classobj.options = dict(DefaultRenderer.options)
            classobj._options_owner = classobj
        # overrid default value from config file
        if dtype is bool:
            value = classobj._config.getbool(name, default)
        else:
            value = dtype(classobj._config.get(name, default))
        # insert into dict
        classobj.options[name] = (value, default, dtype, doc)

    def getOption(self, name):
        """gets the current setting for the option."""
        return self.options[name][0]

    def setOption(self, name, value):
        # make an instance-specific copy of the current option set
        if self._options_owner is not self:
            self.options = dict(self.options)
            self._options_owner = self
        self.options[name][0] = value
        self._config.set(name, value)

    # provide default implementations of rendering methods

    def renderLink(self, relpath=""):
        """renderDefault() is called to render a link to the DP inline (e.g. embedded in text).
        relpath specifies a relative path to the DP location (relative to the HTML page being
        rendered), so should be added to all links. It is guaranteed to end with an "/" if not
        empty.
        Return value should be aa valid HTML string."""
        return renderDefault(self.dp, relpath)

    def renderThumbnail(self, relpath=""):
        """renderThumbnail() is called to render a thumbnail of the DP (e.g. in Data Product tables).
        relpath is as for renderLink() above.
        Return value should be aa valid HTML string, or None or "" if thumbnails are not available."""
        return ""

    def renderLinkComment(self, relpath=""):
        """renderLinkComment() is called to render the DP in link:comment form, as a separate paragraph.
        relpath is as for renderLink() above.
        Return value should be a valid HTML string."""
        items = []
        link = self.renderLink(relpath) or ""
        if link:
            items.append(link)
        comment = self.dp.comment or ""
        if comment:
            comment = comment.replace("<", "&lt;").replace(">", "&gt;")
            items.append(comment)
        if items:
            return "<P>" + ": ".join((items)) + "</P>"
        else:
            return ""

    def renderVerbose(self, relpath=""):
        """renderVerbose() is called to render a verbose version of the DP (e.g. as a separate
        paragraph).
        relpath is as for renderLink() above.
        Return value should be a valid HTML string.
        Default version renders as link:comment"""
        return self.renderThumbnail(relpath) or self.renderLinkComment(relpath)

    def renderInTable(self, relpath=""):
        """renderInTable() is called to render a data product in a table
        relpath is as for renderLink() above.
        Return value should be empty, or a valid HTML string (usually delimited by <TR></TR> tags).
        Default implementation renders thumbnail,comment and link in three table cells."""
        thumb = self.renderThumbnail(relpath) or ""
        comment = self.renderLinkComment(relpath) or ""
        if thumb:
            return "\n".join(
                ["    <TR>"] +
                ["      <TD>%s</TD>" % element for element in (thumb, comment)] +
                ["    </TR>\n"])
        else:
            return """
        <TR><TD COLSPAN=2>%s</TD></TR>\n""" % comment


available_renderers = {}  # type: dict
youngest_renderer = 0


def addRenderer(renderer_class, module_name, module_file):
    global available_renderers
    rdrid = renderer_class.renderer_id
    rdrclass, mod = available_renderers.get(rdrid, (None, None))
    if rdrclass:
        raise RuntimeError("renderer '%s' already registered by module '%s'" % (rdrid, mod))
    renderer_class.module_mtime = mtime = os.path.getmtime(module_file)
    global youngest_renderer
    youngest_renderer = max(mtime, youngest_renderer)
    available_renderers[rdrid] = renderer_class, module_name
    dprintf(1, "registered renderer '%s', mtime %s\n", renderer_class.__name__,
            time.strftime("%x %X", time.localtime(renderer_class.module_mtime)))


def numRenderers():
    global available_renderers
    return len(available_renderers)


def getRenderers(filename):
    """For a given DP, returns a list of renderer ids giving
    the renderers that support the source file type"""
    global available_renderers
    renderers = []
    for rdrid, (renderer, module) in list(available_renderers.items()):
        try:
            priority = renderer.canRender(filename)
        except:
            print(("""Error in renderer: %s.canRender("%s"):""" % (rdrid, filename)))
            traceback.print_exc()
            priority = None
        if priority:
            renderers.append((priority, rdrid))
    # sort by priority
    from functools import cmp_to_key
    renderers.sort(key=cmp_to_key(lambda a, b: cmp(a[0], b[0])))
    # return list of IDs. Note that "none" should always be available and working
    return [a[1] for a in renderers] or ["link"]


def makeRenderer(rdrid, dp, refresh=False):
    """Creates a renderer object with the given rdrid, and attaches it the given DP.
    Failing that, creates a default renderer."""
    try:
        return available_renderers[rdrid][0](dp, refresh=refresh)
    except:
        print(("""Error creating renderer %s for %s:""" % (rdrid, dp.fullpath)))
        traceback.print_exc()
        return DefaultRenderer(dp)


def _callRender(renderer, method, relpath, fallback="%s"):
    try:
        return getattr(renderer, method)(relpath=relpath)
    except:
        dp = getattr(renderer, 'dp', None)
        print(("""Error calling %s.%s for %s""" % (renderer, method, dp and dp.fullpath)))
        traceback.print_exc()
        # if a fallback is specified (and does not contain %s), return that
        if fallback.find("%s") < 0:
            return fallback
        # else render default
        if dp:
            return fallback % renderDefault(dp, relpath)
        else:
            return "&lt;rendering error&gt;"


def renderLink(renderer, relpath=""):
    """calls renderLink() on the specified renderer. On error, falls back to renderDefault()"""
    return _callRender(renderer, "renderLink", relpath)


def renderThumbnail(renderer, relpath=""):
    """calls renderThumbnail() on the specified renderer. On error, returns empty string"""
    return _callRender(renderer, "renderThumbnail", relpath, fallback="")


def renderVerbose(renderer, relpath=""):
    """calls renderVerbose() on the specified renderer. On error, falls back to renderDefault()"""
    return _callRender(renderer, "renderVerbose", relpath, fallback="<P>%s</P>")


def renderInTable(renderer, relpath=""):
    """calls renderLink() on the specified renderer. On error, falls back to renderDefault()"""
    return _callRender(renderer, "renderInTable", relpath, fallback="<TR><TD>%s</TD></TR>")


addRenderer(DefaultRenderer, "built-in", __file__)

# we don't really need this, but it's better to import it here rather than from a plugin, since
# then we detect errors sooner
from . import CachingRenderer



# -*- coding: utf-8 -*-
import fnmatch
import imp
import os
import os.path
import sys
import traceback

import Purr
import Purr.Render
from Purr import dprintf


def loadPlugins(paths):
    global bad_plugins
    bad_plugins = {}
    # find potential plugin files
    modfiles = set()
    for path in paths:
        files = set(fnmatch.filter(os.listdir(path), "*.py"))
        files.discard('__init__.py')
        modfiles.update([os.path.join(path, file) for file in files])
    dprintf(1, "%d plugins found\n", len(modfiles))
    # try to import them
    for fullpath in modfiles:
        modfile = os.path.basename(fullpath)
        modname, ext = os.path.splitext(modfile)
        # try to find the module via the imp mechanism
        try:
            fp, pathname, desc = imp.find_module(modname, paths)
        except:
            err = sys.exc_info()[1]
            dprintf(1, "Error finding module for plugin %s: %s\n", fullpath, err)
            bad_plugins[modname] = err
            continue
        # try to import the module
        #   setattr(modname,"plugin_mtime",os.path.getmtime(pathname))
        try:
            try:
                imp.acquire_lock()
                module = imp.load_module('Purr.Plugins.%s' % modname, fp, pathname, desc)
            finally:
                imp.release_lock()
                if fp:
                    fp.close()
        except:
            err = sys.exc_info()[1]
            dprintf(0, "Error importing module %s: %s\n", pathname, err)
            traceback.print_exc()
            bad_plugins[modname] = err
            continue
        # ok, we have a module, check for vital properties
        dprintf(1, "Imported plugin module '%s' from %s\n", modname, pathname)
    dprintf(1, "%d renderers now available\n", Purr.Render.numRenderers())


# this collects a list of plugins that have failed to load
bad_plugins = None

if bad_plugins is None:
    loadPlugins(__path__)  # type: ignore  # mypy issue #1422



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from . import chart_object
from . import fill_style
from . import font
from . import legend_doc
from . import line_style
from . import theme
from . import tick_mark
from . import pychart_util
from .pychart_types import *


class Entry(chart_object.T):
    keys = {"line_len": (UnitType, None,
                         "Length of the sample line for line plots. If omitted, it is set to be theme.default_font_size"),
            "rect_size": (UnitType, None,
                          "Size of the sample 'blob' for bar range charts. If omitted, it is set to be 70% of theme.default_size"),
            "tick_mark": (tick_mark.T, None, ""),
            "line_style": (line_style.T, None, ""),
            "fill_style": (fill_style.T, None, ""),
            "label": (StringType, "???", ""),
            }
    __doc__ = legend_doc.doc_entry

    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED

    def label_width(self):
        return font.text_width(" " + self.label)

    def get_line_len(self):
        return self.line_len or theme.default_font_size

    def get_rect_size(self):
        return self.rect_size or theme.default_font_size * 7 / 10.0

    def sample_width(self):
        w = 0
        if self.fill_style != None:
            w += self.get_line_len()
        elif self.line_style != None:
            w += self.get_line_len()
        elif self.tick_mark != None:
            w += self.tick_mark.size
        return w

    def height(self):
        h = font.text_height(self.label)[0]
        return h

    def draw(self, ar, can, x_tick, x_label, y):
        """Draw a legend entry. X_TICK and X_LABEL are the X location \
        (in points) of where the sample and label are drawn."""

        rect_size = self.get_rect_size()
        line_len = self.get_line_len()

        nr_lines = len(self.label.split("\n"))
        text_height = font.text_height(self.label)[0]
        line_height = text_height / float(nr_lines)
        y_center = y + text_height - line_height / 1.5

        if self.fill_style != None:
            can.rectangle(self.line_style, self.fill_style,
                          x_tick, y_center - rect_size / 2.0,
                          x_tick + rect_size,
                          y_center + rect_size / 2.0)
        elif self.line_style != None:
            can.line(self.line_style, x_tick, y_center,
                     x_tick + line_len, y_center)
            if self.tick_mark != None:
                self.tick_mark.draw(can, x_tick + line_len / 2.0, y_center)
        elif self.tick_mark != None:
            self.tick_mark.draw(can, x_tick, y_center)

        can.show(x_label, y, self.label)


__doc__ = """Legend is a rectangular box drawn in a chart to describe
the meanings of plots. The contents of a legend box is extracted from
plots' "label", "line-style", and "tick-mark" attributes.

This module exports a single class, legend.T.  Legend.T is a part of
an area.T object, and is drawn automatically when area.draw() method
is called. """


class T(chart_object.T):
    __doc__ = legend_doc.doc
    keys = {
        "inter_row_sep": (UnitType, 0,
                          "Space between each row in the legend."),
        "inter_col_sep": (UnitType, 0,
                          "Space between each column in the legend."),
        "frame_line_style": (line_style.T, line_style.default, ""),
        "frame_fill_style": (fill_style.T, fill_style.white, ""),
        "top_fudge": (UnitType, 0,
                      "Amount of space above the first line."),
        "bottom_fudge": (UnitType, 3,
                         "Amount of space below the last line."),
        "left_fudge": (UnitType, 5,
                       "Amount of space left of the legend."),
        "right_fudge": (UnitType, 5,
                        "Amount of space right of the legend."),
        "loc": (CoordType, None,
                """Bottom-left corner of the legend.
                The default location of a legend is the bottom-right end of the chart."""),
        "shadow": (ShadowType, None, pychart_util.shadow_desc),
        "nr_rows": (IntType, 9999,
                    "Number of rows in the legend. If the number of plots in a chart is larger than nr_rows, multiple columns are created in the legend."),

    }

    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED
    def draw(self, ar, entries, can):
        if not self.loc:
            x = ar.loc[0] + ar.size[0] * 1.1
            y = ar.loc[1]
        else:
            x = self.loc[0]
            y = self.loc[1]

        nr_rows = min(self.nr_rows, len(entries))
        nr_cols = (len(entries) - 1) / nr_rows + 1

        ymin = y
        max_label_width = [0] * nr_cols
        max_sample_width = [0] * nr_cols
        heights = [0] * nr_rows

        for i in range(len(entries)):
            l = entries[i]
            (col, row) = divmod(i, nr_rows)
            max_label_width[col] = max(l.label_width(), max_label_width[col])
            max_sample_width[col] = max(l.sample_width(), max_sample_width[col])
            heights[row] = max(l.height(), heights[row])

        for h in heights:
            y += h
        y += self.inter_row_sep * (nr_rows - 1)
        ymax = y

        tot_width = self.inter_col_sep * (nr_cols - 1)
        for w in max_label_width:
            tot_width += w
        for w in max_sample_width:
            tot_width += w

        can.rectangle(self.frame_line_style, self.frame_fill_style,
                      x - self.left_fudge,
                      ymin - self.bottom_fudge,
                      x + tot_width + self.right_fudge,
                      ymax + self.top_fudge,
                      self.shadow)

        for col in range(nr_cols):
            this_y = y
            this_x = x
            for row in range(nr_rows):
                idx = col * nr_rows + row
                if idx >= len(entries):
                    continue
                this_y -= heights[row]
                l = entries[idx]
                if row != 0:
                    this_y -= self.inter_row_sep

                l.draw(ar, can, this_x, this_x + max_sample_width[col], this_y)
            x += max_label_width[col] + max_sample_width[col] + self.inter_col_sep



# -*- coding: utf-8 -*-
import glob
import os
import os.path

import Kittens.utils
from PyQt4.Qt import QWidget, QDialog, QWizard, QWizardPage, QButtonGroup, QVBoxLayout, QRadioButton, QObject, SIGNAL, QHBoxLayout, QLineEdit, QPushButton, QFileDialog, QMessageBox

import Purr
from Purr import pixmaps


class Error(RuntimeError):
    def __init__(self, message):
        RuntimeError.__init__(self)
        self.error_message = message


wizard_dialog = None


def startWizard(args, mainwin, modal=True):
    """
    Parses list of directories ('args') and attaches to purrlogs appropriately.
    'mainwin' is a Purr.MainWindow object.

    Return value:
      * if modal=True: True if arguments parse OK and Purr.MainWindow should be shown, False on error.
      * if modal=False: True if attached to purrlog and Purr.MainWindow should be shown, False otherwise.

    If modal=False, it will keep displaying the startup dialog as a modeless dialog, and attach the
  mainwin to the specified purrlog when the dialog is closed.

    Use cases:
     $ purr <dirname> [<more dirs>]
        1. If dirname does not exist:
          * find other purrlogs in parent dir
          * pop up wizard, with the name as the default "create" option, and create selected
        2. If dirname exists:
          2a. Not a directory: report error
          2b. If dir is a purrlog, attach and add dirs
          2c. If dir contains 1 valid purrlog, attach, and add more dirs.
          2d. If dir contains 2+ purrlogs, offer choice (or create new), add more dirs
          2e. If dir contains no purrlogs, offer to create one (or to attach to one), add "more dirs" to watchlist if creating new, ignore if attaching
     $ purr
         * same as "purr ."
     $ from meqbrowser, when cd into a directory:
         * if purr dialog is visible, do same as "purr ."
     $ from meqbrowser, when purr button is pressed and purr is not yet visible
         * do "purr ."
     $ from Cattery.Calico, when an MS is selected:
         * do purr MS.purrlog .
    """

    args = args or [os.getcwd()]
    dirname = os.path.abspath(args[0])
    moredirs = args[1:]

    # case 1: dirname does not exist, or refers to a non-directory
    if not os.path.exists(dirname) or not os.path.isdir(dirname):
        message = """To begin with, PURR must load an existing purrlog, or start a new purrlog. <tt>%s</tt> does not look to be an existing purrlog.
            What would you like to do?""" % Kittens.utils.collapseuser(dirname)
        create = dirname
        dirname = os.getcwd()
        parent = os.path.dirname(os.path.normpath(create)) or os.getcwd()
        # if parent is valid dir, find purrlogs in parent (to offer as an option)
        if os.path.isdir(parent):
            purrlogs = list(filter(Purr.Purrer.is_purrlog, glob.glob(os.path.join(parent, "*"))))
        # else use "." as dirname, and do not offer any purrlogs
        else:
            purrlogs = []
    # case 2: dirname exists:
    else:
        create = None
        # case 2b: is a valid purrlog
        if Purr.Purrer.is_purrlog(dirname):
            mainwin.attachPurrlog(dirname, moredirs)
            mainwin.show()
            return True
        # case 2c-2e. Look for purrlogs in dirname
        purrlogs = list(filter(Purr.Purrer.is_purrlog, glob.glob(os.path.join(dirname, "*"))))
        # case 2c: exactly one purrlog. Attach without asking.
        if len(purrlogs) == 1:
            mainwin.show()
            mainwin.attachPurrlog(purrlogs[0], moredirs)
            return True
        # else setup messages
        if purrlogs:
            message = """To begin with, PURR must load an existing purrlog, or start a new purrlog. The directory <tt>%s</tt> contains
          several purrlogs. What would you like to do?""" % Kittens.utils.collapseuser(dirname)
        else:
            message = """To begin with, PURR must load an existing purrlog, or create a new purrlog. The directory <tt>%s</tt> contains
          no purrlogs. What would you like to do?""" % Kittens.utils.collapseuser(dirname)

    # case 1, 2d or 2e: make wizard dialog

    # kill old wizard, if showing
    global wizard_dialog
    if wizard_dialog:
        wizard_dialog.hide()
        dum = QWidget()
        wizard_dialog.setParent(dum)
        dum = wizard_dialog = None

    # create new wizard
    wizard_dialog = PurrStartupWizard(mainwin, dirname, purrlogs, moredirs=moredirs, create=create, message=message)

    if modal:
        if wizard_dialog.exec_() == QDialog.Rejected:
            return False
        return True;
    else:
        wizard_dialog.setModal(False)
        wizard_dialog.show()
        return False


class PurrStartupWizard(QWizard):
    class StartPage(QWizardPage):
        def __init__(self, dirname, purrlogs, parent=None, create=None, message=None):
            QWizardPage.__init__(self, parent)
            self.dirname = dirname
            self.purrlogs = purrlogs or []
            bg = QButtonGroup(self)
            lo = QVBoxLayout()
            self.setLayout(lo)
            # set page titles
            self.setTitle("Starting PURR")
            message and self.setSubTitle(message)
            if not purrlogs:
                self.rbs_log = []
            else:
                # add options for existing purrlogs
                self.rbs_log = [QRadioButton("Load %s" % Kittens.utils.collapseuser(log)) for log in purrlogs]
                for rb in self.rbs_log:
                    lo.addWidget(rb)
                    bg.addButton(rb)
                    QObject.connect(rb, SIGNAL("toggled(bool)"), self.checkCompleteness)
                self.rbs_log[0].setChecked(True)
            # add option to load another purrlog
            lo1 = QHBoxLayout()
            self.rb_other = QRadioButton("Load purrlog from:")
            lo1.addWidget(self.rb_other)
            bg.addButton(self.rb_other)
            self.wother = QLineEdit()
            self.wother.setReadOnly(True)
            lo1.addWidget(self.wother, 1)
            pb = QPushButton(pixmaps.folder_open.icon(), "")
            QObject.connect(pb, SIGNAL("clicked()"), self._select_other_dialog)
            QObject.connect(self.rb_other, SIGNAL("toggled(bool)"), pb.setEnabled)
            QObject.connect(self.rb_other, SIGNAL("toggled(bool)"), self.wother.setEnabled)
            QObject.connect(self.rb_other, SIGNAL("toggled(bool)"), self.checkCompleteness)
            pb.setEnabled(False)
            self.wother.setEnabled(False)
            lo1.addWidget(pb)
            lo.addLayout(lo1)
            self.load_path = None

            # add option to create new purrlog
            lo1 = QHBoxLayout()
            self.rb_create = QRadioButton("Create new purrlog:")
            lo1.addWidget(self.rb_create)
            bg.addButton(self.rb_create)
            self.wcreate = QLineEdit()
            lo1.addWidget(self.wcreate, 1)
            pb = QPushButton(pixmaps.folder_open.icon(), "")
            QObject.connect(pb, SIGNAL("clicked()"), self._select_create_dialog)
            QObject.connect(self.rb_create, SIGNAL("toggled(bool)"), pb.setEnabled)
            QObject.connect(self.rb_create, SIGNAL("toggled(bool)"), self.wcreate.setEnabled)
            QObject.connect(self.rb_create, SIGNAL("toggled(bool)"), self.checkCompleteness)
            QObject.connect(self.wcreate, SIGNAL("editingFinished()"), self._validate_create_filename)
            pb.setEnabled(False)
            self.wcreate.setEnabled(False)
            # this holds the last validated name
            self._validated_create_path = None
            self._validated_result = False
            # generate default name for a new purrlog
            self.create_path = os.path.join(dirname, "purrlog")
            num = 0
            while os.path.exists(self.create_path):
                self.create_path = os.path.join(dirname, "purrlog.%d" % num)
                num += 1
            # This will be not None as long as a valid name is entered
            self.create_path = Kittens.utils.collapseuser(os.path.normpath(self.create_path))
            if create:
                self.wcreate.setText(create or Kittens.utils.collapseuser(create))
                # this will emit checkCompleteness(), causing a _validate_create_filename() call, causing the content of the wcreate widget
                # to be validated and copied to create_path if valid, or reset from create_path if invalid
                self.rb_create.setChecked(True)
            else:
                self.wcreate.setText(self.create_path)

            lo1.addWidget(pb)
            lo.addLayout(lo1)

            # make create option default, if no purrlogs
            if not purrlogs:
                self.rb_create.setChecked(True)

        def _select_other_dialog(self):
            path = str(QFileDialog.getExistingDirectory(self, "Select purrlog", self.dirname))
            if not path:
                return
            if not Purr.Purrer.is_purrlog(path):
                QMessageBox.warning(self, "Invalid purrlog",
                                    "The path you have selected, <tt>%s</tt>, does not refer to a valid purrlog." % Kittens.utils.collapseuser(
                                        path))
                return
            self.load_path = path
            self.wother.setText(Kittens.utils.collapseuser(path))
            self.checkCompleteness()

        def _validate_create_filename(self, path=None, check=True):
            if path is None:
                path = str(self.wcreate.text())
            # if we have already validated this path, then return the last validation result.
            # This is mostly to keep from bombarding the user with repeated error dialogs.
            if self._validated_create_path == path:
                return self._validated_result
            self._validated_create_path = path
            self._validated_result = False;  # set to True if all checks pass
            # now process the path. Normalize it, and expand "~"
            path = os.path.expanduser(os.path.normpath(path))
            # if not absolute, join to current directory
            if not os.path.isabs(path):
                path = os.path.join(self.dirname, path)
            # collapse to "~" (for error messages)
            path0 = Kittens.utils.collapseuser(path)
            if os.path.exists(path):
                QMessageBox.warning(self, "Can't create purrlog",
                                    """Unable to create purrlog <tt>%s</tt>: file or directory already exists. Please select another name""" % path0)
                self.create_path and self.wcreate.setText(Kittens.utils.collapseuser(self.create_path))
                return False
            if not os.access(os.path.dirname(os.path.normpath(path)) or '.', os.W_OK):
                QMessageBox.warning(self, "Can't create purrlog",
                                    """Unable to create purrlog <tt>%s</tt>: can't write to parent directory. Please select another path.""" % path0)
                self.create_path and self.wcreate.setText(Kittens.utils.collapseuser(self.create_path))
                return False
            self.create_path = path
            self.wcreate.setText(path0)
            self._validated_result = True;  # set to True if all checks pass
            if check:
                self.checkCompleteness()
            return True

        def _select_create_dialog(self):
            path = str(QFileDialog.getSaveFileName(self, "Create new purrlog", self.create_path))
            if path:
                self._validate_create_filename(path)

        def checkCompleteness(self, toggled=None):
            if toggled and hasattr(self, 'rb_other') and self.rb_other.isChecked() and not self.load_path:
                self._select_other_dialog()
            else:
                self.emit(SIGNAL("completeChanged()"))

        def isComplete(self):
            if hasattr(self, 'rb_create') and self.rb_create.isChecked():
                return self._validate_create_filename(check=False) and bool(self.create_path)
            if hasattr(self, 'rb_other') and self.rb_other.isChecked():
                return bool(self.load_path)
            return True

        def selectedPath(self):
            for (rb, log) in zip(self.rbs_log, self.purrlogs):
                if rb.isChecked():
                    return log
            if self.rb_other.isChecked():
                return self.load_path
            if self.rb_create.isChecked():
                return self.create_path
            return None

    def __init__(self, mainwin, dirname, purrlogs, moredirs=[], create=None, message=None):
        QWizard.__init__(self, mainwin)
        self.setWindowTitle("Starting PURR")
        self.setPixmap(QWizard.LogoPixmap, pixmaps.purr_logo.pm())
        self.setOption(QWizard.NoBackButtonOnStartPage)
        self.setButtonText(QWizard.FinishButton, "Proceed")
        # create start page
        self._startpage = self.StartPage(dirname, purrlogs, create=create, message=message)
        self.addPage(self._startpage)
        # internal state
        self._dirname = dirname
        self._mainwin = mainwin
        self._moredirs = moredirs

    def done(self, code):
        if code == QDialog.Accepted:
            # check path, set code to rejected if none set
            path = self._startpage.selectedPath()
            if not path:
                print("No path selected in StartupWizard. This is probably a bug, please report it!")
                code = QDialog.Rejected
            else:
                # show the main window
                self._mainwin.show()
                # if attaching to existing purrlog, cancel the moredirs argument
                # if creating new purrlog, add parent directory to watchlist
                moredirs = None if os.path.exists(path) else [self._dirname] + list(self._moredirs)
                self._mainwin.attachPurrlog(path, moredirs)
        return QDialog.done(self, code)

    def selectedPath(self):
        return self._startpage.selectedPath()

# class PurrlogSelectWizard (QDialog):
#  def __init__(self,parent,dirname,purrlogs):
#    QDialog.__init__(self,parent)
#    self.setWindowTitle("PURR Startup Wizard")
#    self._currier = Kittens.utils.PersistentCurrier()
#    self._lo = QVBoxLayout(self)
#    self._lo.setSpacing(0)
#    for log in purrlogs:
#      self.button(pixmaps.purr_logo.icon(),"Load %s"%Kittens.utils.collapseuser(log),self._currier.curry(self._load_log,log))
#    self.button(pixmaps.purr_logo.icon(),"Load a different purrlog...",self._load_other_log)
#    self.button(pixmaps.purr_logo.icon(),"Create new purrlog in %s..."%Kittens.utils.collapseuser(dirname),self._create_log)
#    self._lo.addSpacing(10)
#    self.button(pixmaps.red_round_cross.icon(),"Cancel",self.reject)
#
#  def button (self,icon,text,callback):
#    pb = QToolButton(self)
#    pb.setIcon(icon)
#    pb.setText(text)
#    pb.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
#    pb.setSizePolicy(QSizePolicy.Minimum,QSizePolicy.MinimumExpanding)
#    # pb.setFlat(True)
#    self._lo.addWidget(pb)
#    QObject.connect(pb,SIGNAL("clicked()"),callback)
#    return pb
#
#  def _load_log (self,logname):
#    pass
#
#  def _create_log (self):
#    pass
#
#  def _load_other_log (self):
#    dialog = QFileDialog(parent,"Open or create purrlog",dirname,"*purrlog")
#    try:
#      dialog.setOption(QFileDialog.ShowDirsOnly)
#      dialog.setFileMode(QFileDialog.Directory)
#    except AttributeError: # Qt 4.4 has no setOption
#      dialog.setFileMode(QFileDialog.DirectoryOnly)
#    if purrlogs:
#      dialog.setSidebarUrls(map(QUrl,purrlogs))
#    if not dialog.exec_():
#      return False
#    logname = str(dialog.selectedFiles()[0])
#    mainwin.show()
#    mainwin.attachPurrlog(logname,moredirs)
#    return True



# -*- coding: utf-8 -*-
try:
    import PIL.Image
except:
    print("""Please install the Python Imaging Library (PIL)!
Purr will work fine without it, but your logs will be a lot less beautifully rendered.
PIL is available from http://www.pythonware.com/products/pil/. On Debian-based systems
(including Ubuntu and such), it can be as simple as installing the python-imaging package.""")
    raise

import os.path
import traceback

from Purr.Render import DefaultRenderer, quote_url


class ImageRenderer(DefaultRenderer):
    """This class renders PIL-compatible image data products."""
    _extensions = set([
        "jpg", "jpeg", "png", "gif", "xpm", "ppm", "pbm", "pnm", "tiff", "tif"
    ])

    @staticmethod
    def canRender(filename):
        """Check extensions."""
        name, ext = os.path.splitext(filename)
        ext = ext.lstrip('.').lower()
        if ext in ImageRenderer._extensions:
            return 100
        else:
            return False

    # this gives a short ID for the class (used in GUIs and such)
    renderer_id = "image"

    # this gives a documentation string. You can use rich text here
    renderer_doc = """<P>The "image" plugin provides rendering of image-type data products.
      It can render any image that is compatible with the Python Image Library."""

    # define renderer options
    DefaultRenderer.addOption("image-thumbnail-width", 512, dtype=int, doc="Maximum width of thumbnails")
    DefaultRenderer.addOption("image-thumbnail-height", 256, dtype=int, doc="Maximum height of thumbnails")

    def __init__(self, dp, **kw):
        DefaultRenderer.__init__(self, dp, **kw)
        Purr.progressMessage("rendering %s" % dp.filename, sub=True)
        img = PIL.Image.open(dp.fullpath)
        # If image format is not compatible with browsers, save it in PNG format.
        # self.fullimage will refer to the browser-compatible image, to which the thumbnail will link
        if img.format not in ["GIF", "JPEG", "PNG"]:
            self.fullimage, path, uptodate = self.subproduct("-full.png")
            if not uptodate:
                try:
                    img.save(path, "PNG")
                except:
                    print(("Error saving %s in PNG format" % path))
                    traceback.print_exc()
                    self.fullimage = dp.filename
        else:
            self.fullimage = dp.filename
        # now, do we need to generate a thumbnail?
        tsize = self.getOption("image-thumbnail-width"), self.getOption("image-thumbnail-height")
        width, height = img.size
        factors = width / float(tsize[0]), height / float(tsize[1])
        if max(factors) <= 1:
            self.thumbnail = self.fullimage
        else:
            # regenerate thumbnail if needed
            self.thumbnail, path, uptodate = self.subproduct("-thumb.png")
            try:
                # regenerate only if doesn't exist
                if not uptodate:
                    # use max resize factor
                    factor = max(factors)
                    img = img.resize((int(width / factor), int(height / factor)), PIL.Image.ANTIALIAS)
                    img.save(path, "PNG")
            except:
                print(("Error saving thumbnail %s in PNG format" % path))
                traceback.print_exc()
                self.thumbnail = None

    def renderThumbnail(self, relpath=""):
        """renderThumbnail() is called to render a thumbnail of the DP (e.g. in Data Product tables)."""
        # no thumbnail -- return empty string
        if self.thumbnail is None:
            return ""
        # else thumbnail is same as full image (because image was small enough), insert directly
        elif self.thumbnail is self.fullimage:
            fname = relpath + self.fullimage
            return """<IMG SRC="%s" ALT="%s"></IMG>""" % (quote_url(fname), quote_url(os.path.basename(self.filename)))
        # else return thumbnail linking to full image
        else:
            tname = relpath + self.thumbnail
            fname = relpath + self.fullimage
            return """<A HREF="%s"><IMG SRC="%s" ALT="%s"></A>""" % (
            quote_url(fname), quote_url(tname), quote_url(os.path.basename(self.filename)))


# register ourselves with Purr
import Purr.Render

Purr.Render.addRenderer(ImageRenderer, __name__, __file__)
